



if exist %PREFIX%\\Library\\include\\ignition\\transport10\\ignition\\transport.hh (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if exist $PREFIX$\\Library\\lib\\ignition-transport10.lib (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if exist $PREFIX$\\Library\\bin\\ignition-transport10.dll (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if exist %PREFIX%\\Library\\lib\\cmake\\ignition-transport10\\ignition-transport10-config.cmake (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
